@@ -24,7 +24,6 @@
 #include	"numlib/numlib.h"
 #include	"tcpd/spipe.h"
 #include	"tcpd/tlsclient.h"
-#include	<courierauthsaslclient.h>
 
 #include	"rw.h"
 #include	"esmtpconfig.h"
@@ -373,6 +372,17 @@ static void sendesmtp(struct esmtp_info *info, struct my_esmtp_info *my_info)
 	if (esmtp_connect(info, my_info))
 		return;
 
+	/*
+	** If this message has 8 bit headers, this mail server better have
+	** smtputf8 support.
+	*/
+	if (ctlfile_searchfirst(ctf, COMCTLFILE_SMTPUTF8) >= 0 &&
+	    !info->hassmtputf8)
+	{
+		esmtp_unicode_required_error(info, my_info);
+		return;
+	}
+
 	/*
 	** Ok, we now have a connection.  We want to call push() to deliver
 	** this message, but if the VERP flag is set but the remote server
@@ -834,7 +844,7 @@ static struct esmtp_rcpt_info *mk_rcpt_info(struct moduledel *del,
 
 		p[i].address=del->receipients[i*2+1];
 		p[i].dsn_options=ctf->dsnreceipients[n];
-		p[i].orig_receipient=ctf->oreceipients[n];
+		p[i].orig_receipient=ctf->oreceipients_utf8[n];
 	}
 
 	return p;
