@@ -1,5 +1,5 @@
 /*
-** Copyright 2017 Double Precision, Inc.
+** Copyright 2017-2018 Double Precision, Inc.
 ** See COPYING for distribution information.
 */
 
@@ -8,6 +8,7 @@
 #endif
 #include	"libesmtp.h"
 #include	"smtproutes.h"
+#include	"comsts.h"
 #include	<errno.h>
 #include	<sys/types.h>
 #include	<sys/uio.h>
@@ -20,6 +21,7 @@
 #include	"tcpd/tlsclient.h"
 #include	"rfc2045/rfc2045.h"
 #include	"rw.h"
+#include	<idna.h>
 #include	<courierauthsaslclient.h>
 
 static void esmtp_wait_rw(struct esmtp_info *info, int *waitr, int *waitw);
@@ -37,14 +39,13 @@ static int esmtp_helo(struct esmtp_info *info, int using_tls,
 static int esmtp_get_greeting(struct esmtp_info *info,
 			      void *arg);
 static int esmtp_enable_tls(struct esmtp_info *,
+			    enum sts_mode,
 			    const char *,
 			    int,
 			    void *arg);
 static int esmtp_auth(struct esmtp_info *info,
 		      const char *auth_key,
 		      void *arg);
-static char *esmtp_mailfrom_cmd(struct esmtp_info *info,
-				struct esmtp_mailfrom_info *mf_info);
 
 static int esmtp_sendcommand(struct esmtp_info *info,
 			     const char *cmd,
@@ -53,6 +54,15 @@ static int esmtp_parsereply(struct esmtp_info *info,
 			    const char *cmd,
 			    void *arg);
 
+static const char * const libesmtp_unicode_err[]=
+	{
+	 "553-The recipient's mail server does not support E-mail",
+	 "553-messages with international Unicode E-mail addresses or",
+	 "553-E-mail messages in the international Unicode E-mail"
+	 " format.",
+	 "553 Unable to deliver Unicode E-mail to a non-Unicode mail server.",
+	 0,
+	};
 static void connect_error(struct esmtp_info *info, void *arg)
 {
 	(*info->log_smtp_error)(info,
@@ -559,12 +569,14 @@ static int esmtp_helo(struct esmtp_info *info, int using_tls,
 	info->haspipelining=0;
 	info->hasdsn=0;
 	info->has8bitmime=0;
+	info->hassmtputf8=0;
 	info->hasverp=0;
 	info->hassize=0;
 	info->hasexdata=0;
 	info->hascourier=0;
 	info->hasstarttls=0;
 	info->hassecurity_starttls=0;
+	info->is_tls_connection=0;
 	info->is_secure_connection=0;
 
 
@@ -608,7 +620,20 @@ static int esmtp_helo(struct esmtp_info *info, int using_tls,
 
 	esmtp_timeout(info, info->helo_timeout);
 	strcpy(hellobuf, "EHLO ");
-	strncat(hellobuf, p, sizeof(hellobuf)-10);
+
+	{
+		char *q;
+
+		if (idna_to_ascii_8z(p, &q, 0) == IDNA_SUCCESS)
+		{
+			strncat(hellobuf, q, sizeof(hellobuf)-10);
+			free(q);
+		}
+		else
+		{
+			strncat(hellobuf, p, sizeof(hellobuf)-10);
+		}
+	}
 	strcat(hellobuf, "\r\n");
 
 	if (esmtp_writestr(info, hellobuf) || esmtp_writeflush(info))
@@ -722,6 +747,8 @@ static int esmtp_helo(struct esmtp_info *info, int using_tls,
 				info->hasdsn=1;
 			if (strcmp(hellobuf, "8BITMIME") == 0)
 				info->has8bitmime=1;
+			if (strcmp(hellobuf, "SMTPUTF8") == 0)
+				info->hassmtputf8=1;
 			if (strcmp(hellobuf, "SIZE") == 0)
 				info->hassize=1;
 			if (strcmp(hellobuf, "STARTTLS") == 0)
@@ -828,9 +855,31 @@ static int esmtp_helo(struct esmtp_info *info, int using_tls,
 
 	if (getenv("COURIER_ESMTP_DEBUG_NO8BITMIME"))
 		info->has8bitmime=0;
+
+	if (getenv("COURIER_ESMTP_DEBUG_NOSMTPUTF8"))
+		info->hassmtputf8=0;
+
+	if (getenv("COURIER_ESMTP_DEBUG_NOPIPELINING"))
+		info->haspipelining=0;
 	return (0);
 }
 
+void esmtp_unicode_required_error(struct esmtp_info *info,
+				  void *arg)
+{
+	const char * const *p=libesmtp_unicode_err;
+
+	(*info->log_talking)(info, arg);
+
+	while (p[1])
+	{
+		(*info->log_reply)(info, *p, -1, arg);
+		++p;
+	}
+
+	(*info->log_smtp_error)(info, *p, **p, arg);
+}
+
 static void connection_closed(struct esmtp_info *info,
 			      void *arg)
 {
@@ -841,6 +890,7 @@ static void connection_closed(struct esmtp_info *info,
 }
 
 static int esmtp_enable_tls(struct esmtp_info *info,
+			    enum sts_mode domain_sts_mode,
 			    const char *hostname, int smtps,
 			    void *arg)
 {
@@ -853,12 +903,13 @@ static int esmtp_enable_tls(struct esmtp_info *info,
 	char	remotefd_buf[NUMBUFSIZE+30];
 	char	miscbuf[NUMBUFSIZE];
 
-	static char *trustcert_buf=0;
-	static char *origcert_buf=0;
+	char *trustcert_buf=0;
+
+	char *vars[10];
+	int vars_n=0;
 
 	char *argvec[10];
 
-	int restore_origcert=0;
 	int n;
 
 	if (libmail_streampipe(pipefd))
@@ -918,27 +969,18 @@ static int esmtp_enable_tls(struct esmtp_info *info,
 
 	p=getenv("ESMTP_TLS_VERIFY_DOMAIN");
 
+	if (domain_sts_mode == sts_mode_enforce)
+		p="1";
+
 	if ((info->smtproutes_flags & ROUTE_STARTTLS) != 0)
 	{
-		char *q, *r;
+		char *q;
 
 		/*
 		** Replace TLS_TRUSTCERTS with TLS_TRUSTSECURITYCERTS,
 		** until couriertls is execed.
 		*/
 
-		q=getenv("TLS_TRUSTCERTS");
-
-		r=malloc(strlen(q ? q:"")+40);
-		if (!r)
-			abort();
-		strcat(strcpy(r, "TLS_TRUSTCERTS="), q ? q:"");
-
-		if (origcert_buf)
-			free(origcert_buf);
-		origcert_buf=r;
-		restore_origcert=1;
-
 		p=getenv("TLS_TRUSTSECURITYCERTS");
 		if (!p || !*p)
 		{
@@ -960,12 +1002,8 @@ static int esmtp_enable_tls(struct esmtp_info *info,
 		if (!q)
 			abort();
 		strcat(strcpy(q, "TLS_TRUSTCERTS="), p);
-		putenv(q);
+		vars[vars_n++]=q;
 		p="1";
-
-		if (trustcert_buf)
-			free(trustcert_buf);
-		trustcert_buf=q;
 	}
 
 	if (p && atoi(p))
@@ -987,10 +1025,22 @@ static int esmtp_enable_tls(struct esmtp_info *info,
 	}
 	argvec[n]=0;
 
+	if (trustcert_buf)
+	{
+		vars[vars_n++]=trustcert_buf;
+	}
+
+	if (domain_sts_mode == sts_mode_enforce)
+	{
+		vars[vars_n++]="TLS_VERIFYPEER=PEER";
+	}
+	vars[vars_n]=0;
+	cinfo.override_vars=vars;
+
 	n=couriertls_start(argvec, &cinfo);
 
-	if (restore_origcert)
-		putenv(origcert_buf);
+	if (trustcert_buf)
+		free(trustcert_buf);
 	if (verify_domain)
 		free(verify_domain);
 
@@ -1027,6 +1077,8 @@ static int esmtp_enable_tls(struct esmtp_info *info,
 
 	esmtp_init(info);
 
+	info->is_tls_connection=1;
+
 	/* Ask again for an EHLO, because the capabilities may differ now */
 
 	if (smtps)
@@ -1364,23 +1416,52 @@ static int local_sock_address(struct esmtp_info *info,
 	return 0;
 }
 
+static int do_esmtp_connect_to(struct esmtp_info *info,
+			       const char *connect_to,
+			       struct sts_id *id,
+			       enum sts_mode domain_sts_mode,
+			       void *arg);
+
+
 static int do_esmtp_connect(struct esmtp_info *info, void *arg)
+{
+	const char *dest_domain=info->smtproute ? info->smtproute:info->host;
+	struct sts_id domain_sts_id;
+	int rc;
+	const char *tls_trustcerts=getenv("TLS_TRUSTCERTS");
+
+	/* If TLS is not enabled, disable STS */
+
+	sts_init(&domain_sts_id, (tls_trustcerts && *tls_trustcerts
+				  ? dest_domain:NULL));
+
+	rc=do_esmtp_connect_to(info, dest_domain, &domain_sts_id,
+			       get_sts_mode(&domain_sts_id),
+			       arg);
+
+	sts_deinit(&domain_sts_id);
+
+	return rc;
+}
+
+static int do_esmtp_connect_to(struct esmtp_info *info,
+			       const char *connect_to,
+			       struct sts_id *domain_sts_id,
+			       enum sts_mode domain_sts_mode,
+			       void *arg)
 {
 	struct rfc1035_mxlist *mxlist, *p, *q;
 	int static_route= info->smtproute != NULL;
 	struct rfc1035_res res;
 	int rc;
-	const char *auth_key;
 	int connection_attempt_made;
 
 	errno=0;	/* Detect network failures */
 
-	auth_key=info->smtproute ? info->smtproute:info->host;
-
 	rfc1035_init_resolv(&res);
 
 	rc=rfc1035_mxlist_create_x(&res,
-				   auth_key,
+				   connect_to,
 				   RFC1035_MX_AFALLBACK |
 				   RFC1035_MX_IGNORESOFTERR,
 				   &mxlist);
@@ -1476,6 +1557,10 @@ static int do_esmtp_connect(struct esmtp_info *info, void *arg)
 					 sizeof(p->address), &port))
 			continue;
 
+		if (domain_sts_mode == sts_mode_enforce &&
+		    sts_mx_validate(domain_sts_id, p->hostname) < 0)
+			continue;
+
 		connection_attempt_made=1;
 
 		info->sockfdaddr=addr;
@@ -1510,6 +1595,7 @@ static int do_esmtp_connect(struct esmtp_info *info, void *arg)
 			if (info->smtproutes_flags & ROUTE_SMTPS)
 			{
 				if (esmtp_enable_tls(info,
+						     domain_sts_mode,
 						     p->hostname,
 						     1,
 						     arg))
@@ -1551,11 +1637,17 @@ static int do_esmtp_connect(struct esmtp_info *info, void *arg)
 						(info,
 						 "/SECURITY set, but TLS is not available",
 						 '5', arg);
-					}
+				}
+				else if (domain_sts_mode == sts_mode_enforce
+					 && !info->hasstarttls)
+				{
+					rc=1;
+				}
 				else if (info->hasstarttls &&
-					 esmtp_enable_tls
-					 (info, p->hostname,
-					  0, arg))
+					 esmtp_enable_tls(info,
+							  domain_sts_mode,
+							  p->hostname,
+							  0, arg))
 				{
 					/*
 					** Only keep track of broken
@@ -1565,7 +1657,8 @@ static int do_esmtp_connect(struct esmtp_info *info, void *arg)
 					** Otherwise, it's on us.
 					*/
 					if (!(info->smtproutes_flags &
-					      ROUTE_STARTTLS))
+					      ROUTE_STARTTLS) &&
+					    domain_sts_mode != sts_mode_enforce)
 					{
 						(*info->report_broken_starttls)
 							(info,
@@ -1575,7 +1668,8 @@ static int do_esmtp_connect(struct esmtp_info *info, void *arg)
 				}
 				else
 				{
-					int rc=esmtp_auth(info, auth_key, arg);
+					int rc=esmtp_auth(info, connect_to,
+							  arg);
 
 					if (rc == 0)
 						break; /* Good HELO */
@@ -1768,29 +1862,75 @@ static int esmtp_parsereply(struct esmtp_info *info,
 	return (0);
 }
 
-
 /*
 ** Construct the MAIL FROM: command, taking into account ESMTP capabilities
 ** of the remote server.
 */
+static char *esmtp_mailfrom_cmd_idna(struct esmtp_info *info,
+				     const char *sender,
+				     struct esmtp_mailfrom_info *mf_info,
+				     const char **errmsg);
 
-static char *esmtp_mailfrom_cmd(struct esmtp_info *info,
-				struct esmtp_mailfrom_info *mf_info)
+char *esmtp_mailfrom_cmd(struct esmtp_info *info,
+			 struct esmtp_mailfrom_info *mf_info,
+			 const char **errmsg)
+{
+	if (!info->hassmtputf8)
+	{
+		char *idna_sender=udomainace(mf_info->sender);
+		char *r=esmtp_mailfrom_cmd_idna(info, idna_sender,
+						mf_info, errmsg);
+
+		free(idna_sender);
+		return r;
+	}
+
+	return esmtp_mailfrom_cmd_idna(info, mf_info->sender, mf_info, errmsg);
+}
+
+static char *esmtp_mailfrom_cmd_idna(struct esmtp_info *info,
+				     const char *sender,
+				     struct esmtp_mailfrom_info *mf_info,
+				     const char **errmsg)
 {
 	char	*bodyverb="", *verpverb="", *retverb="";
 	char	*oenvidverb="", *sizeverb="";
+	char	*smtputf8verb="";
+
 	const char *seclevel="";
 	char	*mailfromcmd;
 	size_t l;
 
 	static const char seclevel_starttls[]=" SECURITY=STARTTLS";
 
+	if (errmsg)
+		*errmsg=0;
+
 	if (info->has8bitmime)	/* ESMTP 8BITMIME capability */
 		bodyverb= mf_info->is8bitmsg ? " BODY=8BITMIME":" BODY=7BIT";
 
 	if (info->hasverp && mf_info->verp)
 		verpverb=" VERP";	/* ESMTP VERP capability */
 
+	if (info->hassmtputf8)
+		smtputf8verb=" SMTPUTF8";
+	else
+	{
+		const char *p;
+
+		for (p=sender; *p; p++)
+			if (*p & 0x80)
+			{
+				if (errmsg)
+					*errmsg="550-The addressee's mail"
+						"server "
+						"does not support the sender's "
+						"international E-mail address.";
+
+				return NULL;
+			}
+	}
+
 	/* ESMTP DSN capability */
 	if (info->hasdsn && mf_info->dsn_format)
 		retverb=mf_info->dsn_format == 'F' ? " RET=FULL":
@@ -1832,12 +1972,13 @@ static char *esmtp_mailfrom_cmd(struct esmtp_info *info,
 		seclevel=seclevel_starttls;
 
 	l=sizeof("MAIL FROM:<>\r\n")+
-		strlen(mf_info->sender)+
+		strlen(sender)+
 		strlen(bodyverb)+
 		strlen(verpverb)+
 		strlen(retverb)+
 		strlen(oenvidverb)+
 		strlen(sizeverb)+
+		strlen(smtputf8verb)+
 		strlen(seclevel);
 
 	mailfromcmd=malloc(l);
@@ -1845,8 +1986,9 @@ static char *esmtp_mailfrom_cmd(struct esmtp_info *info,
 	if (!mailfromcmd)
 		abort();
 
-	snprintf(mailfromcmd, l, "MAIL FROM:<%s>%s%s%s%s%s%s\r\n",
-		 mf_info->sender,
+	snprintf(mailfromcmd, l, "MAIL FROM:<%s>%s%s%s%s%s%s%s\r\n",
+		 sender,
+		 smtputf8verb,
 		 bodyverb,
 		 verpverb,
 		 retverb,
@@ -1859,16 +2001,26 @@ static char *esmtp_mailfrom_cmd(struct esmtp_info *info,
 	return (mailfromcmd);
 }
 
+
 /*
 ** Build RCPT TOs
 */
 
+static void mk_one_receip_idna_or_utf8(struct esmtp_info *info,
+				       struct esmtp_rcpt_info *receip,
+				       const char *utf8_or_idna_address,
+				       const char *utf8_or_idna_orig_receip,
+				       void (*builder)(const char *, void *),
+				       void *arg);
+
 static void mk_one_receip(struct esmtp_info *info,
 			  struct esmtp_rcpt_info *receip,
 			  void (*builder)(const char *, void *),
 			  void *arg)
 {
-	const char *p;
+	char *idna_address=0;
+	const char *orig_receip;
+	char *idna_orig_receip=0;
 
 	if (!receip)
 	{
@@ -1876,8 +2028,58 @@ static void mk_one_receip(struct esmtp_info *info,
 		return;
 	}
 
+	orig_receip=0;
+
+	if (info->hasdsn)
+	{
+		const char *p=receip->orig_receipient;
+
+		if (p && *p)
+		    orig_receip=p;
+	}
+
+	if (!info->hassmtputf8)
+	{
+		idna_address=udomainace(receip->address);
+		if (orig_receip)
+			idna_orig_receip=udomainace(orig_receip);
+	}
+
+	mk_one_receip_idna_or_utf8
+		(info, receip,
+		 (idna_address ? idna_address:receip->address),
+		 (idna_orig_receip ? idna_orig_receip:orig_receip),
+		 builder, arg);
+
+	if (idna_address)
+		free(idna_address);
+	if (idna_orig_receip)
+		free(idna_orig_receip);
+}
+
+static void mk_one_receip_idna_or_utf8(struct esmtp_info *info,
+				       struct esmtp_rcpt_info *receip,
+				       const char *utf8_or_idna_address,
+				       const char *utf8_or_idna_orig_receip,
+				       void (*builder)(const char *, void *),
+				       void *arg)
+{
+	const char *p;
+
+	if (!info->hassmtputf8)
+	{
+		const char *p;
+
+		for (p=utf8_or_idna_address; *p; p++)
+			if ( (*p) & 0x80 )
+			{
+				receip->our_rcpt_error=libesmtp_unicode_err;
+				return;
+			}
+	}
+
 	(*builder)("RCPT TO:<", arg);
-	(*builder)(receip->address, arg);
+	(*builder)(utf8_or_idna_address, arg);
 	(*builder)(">", arg);
 
 	p=receip->dsn_options;
@@ -1928,12 +2130,24 @@ static void mk_one_receip(struct esmtp_info *info,
 		}
 	}
 
-	p=receip->orig_receipient;
-
-	if (p && *p && info->hasdsn)
+	if (utf8_or_idna_orig_receip)
 	{
+		char *encoded;
+
+		if (info->hassmtputf8)
+		{
+			encoded=rfc6533_encode(utf8_or_idna_orig_receip, 0);
+		}
+		else
+		{
+			char *p=udomainace(utf8_or_idna_orig_receip);
+
+			encoded=rfc6533_encode(p, 1);
+			free(p);
+		}
 		(*builder)(" ORCPT=", arg);
-		(*builder)(p, arg);
+		(*builder)(encoded, arg);
+		free(encoded);
 	}
 
 	(*builder)("\r\n", arg);
@@ -1968,6 +2182,9 @@ static char *rcpt_data(struct esmtp_info *info,
 	if (append_data)
 		++n;
 
+	for (i=0; i < nreceips; ++i)
+		receips[i].our_rcpt_error=0;
+
 	for (i=0; i < n; ++i)
 		mk_one_receip(info, (i<nreceips ? &receips[i]:NULL),
 			      count, &l);
@@ -2050,6 +2267,7 @@ static int parsedatareply(struct esmtp_info *info,
 
 static int do_pipeline_rcpt_2(struct esmtp_info *info,
 			      int *rcptok,
+			      struct esmtp_rcpt_info *receips,
 			      size_t nrecipients,
 			      char *rcpt_data_cmd,
 			      void *arg)
@@ -2076,6 +2294,21 @@ static int do_pipeline_rcpt_2(struct esmtp_info *info,
 		char *this_rcpt_to_cmd;
 		size_t this_rcpt_to_len;
 
+		if (receips[i].our_rcpt_error)
+		{
+			const char * const *p=receips[i].our_rcpt_error;
+
+			rcptok[i]=0;
+			(*info->log_reply)(info, *p, **p, arg);
+			while (p[1])
+			{
+				++p;
+				(*info->log_reply)(info, *p, **p, arg);
+			}
+			(*info->log_rcpt_error)(info, i, **p, arg);
+			continue;
+		}
+
 		for (this_rcpt_to_len=0; orig_rcpt_data_cmd[this_rcpt_to_len];
 		     ++this_rcpt_to_len)
 			if (orig_rcpt_data_cmd[this_rcpt_to_len] == '\n')
@@ -2604,6 +2837,7 @@ int esmtp_send(struct esmtp_info *info,
 
 	struct	stat stat_buf;
 	char *rcpt_buf;
+	const char *errmsg;
 
 	if (info->esmtp_sockfd < 0)
 		return -1;
@@ -2611,11 +2845,17 @@ int esmtp_send(struct esmtp_info *info,
 	if (fstat(fd, &stat_buf) == 0)
 		mf_info->msgsize=stat_buf.st_size;
 
-	mailfroms=esmtp_mailfrom_cmd(info, mf_info);
+	mailfroms=esmtp_mailfrom_cmd(info, mf_info, &errmsg);
 
 	(*info->log_talking)(info, arg);
 	esmtp_timeout(info, info->cmd_timeout);
 
+	if (!mailfroms)
+	{
+		(*info->log_smtp_error)(info, errmsg, *errmsg, arg);
+		return 0;
+	}
+
 	if (esmtp_sendcommand(info, mailfroms, arg))
 	{
 		free(mailfroms);
@@ -2651,7 +2891,7 @@ int esmtp_send(struct esmtp_info *info,
 
 	rcpt_buf=esmtp_rcpt_data_create(info, rcpt_info, nreceipients);
 
-	if ( do_pipeline_rcpt_2(info, rcptok, nreceipients, rcpt_buf,
+	if ( do_pipeline_rcpt_2(info, rcptok, rcpt_info, nreceipients, rcpt_buf,
 				arg) )
 	{
 		free(rcpt_buf);
